"""Implementations of default dereference actions."""

from .context import ActionContext
from rid_lib.types import KoiNetNode
from rid_lib.ext import Bundle
from .effector import Effector


@Effector.register_default_action(KoiNetNode)
def dereference_koi_node(
    ctx: ActionContext, rid: KoiNetNode
) -> Bundle | None:
    """Dereference function for this KOI node.
    
    Generates a bundle from this node's profile data in the config.
    """
    
    if rid != ctx.identity.rid:
        return
    
    return Bundle.generate(
        rid=ctx.identity.rid,
        contents=ctx.identity.profile.model_dump()
    )


import logging
from rich.logging import RichHandler
from pydantic import Field
from rid_lib.types import KoiNetNode, KoiNetEdge
from koi_net.config import NodeConfig, KoiNetConfig, ServerConfig
from koi_net.core import NodeAssembler
from koi_net.protocol.node import NodeProfile, NodeProvides, NodeType
from koi_net.context import HandlerContext
from koi_net.processor.handler import HandlerType, KnowledgeHandler
from koi_net.processor.knowledge_object import KnowledgeObject
from koi_net.protocol.event import Event, EventType
from koi_net.protocol.edge import EdgeType, generate_edge_bundle

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    handlers=[RichHandler()]
)

logging.getLogger("koi_net").setLevel(logging.DEBUG)
logger = logging.getLogger(__name__)


class CoordinatorConfig(NodeConfig):
    server: ServerConfig = Field(default_factory=lambda: 
        ServerConfig(port=8080)
    )
    koi_net: KoiNetConfig = Field(default_factory = lambda:
        KoiNetConfig(
            node_name="coordinator",
            node_profile=NodeProfile(
                node_type=NodeType.FULL,
                provides=NodeProvides(
                    event=[KoiNetNode, KoiNetEdge],
                    state=[KoiNetNode, KoiNetEdge]
                )
            ),
            rid_types_of_interest=[KoiNetNode, KoiNetEdge]
        )
    )

@KnowledgeHandler.create(
    HandlerType.Network, 
    rid_types=[KoiNetNode])
def handshake_handler(ctx: HandlerContext, kobj: KnowledgeObject):
    logger.info("Handling node handshake")

    # only respond if node declares itself as NEW
    if kobj.event_type != EventType.NEW:
        return
        
    logger.info("Sharing this node's bundle with peer")
    identity_bundle = ctx.cache.read(ctx.identity.rid)
    ctx.event_queue.push_event_to(
        event=Event.from_bundle(EventType.NEW, identity_bundle),
        target=kobj.rid
    )
    
    logger.info("Proposing new edge")    
    # defer handling of proposed edge
    
    edge_bundle = generate_edge_bundle(
        source=kobj.rid,
        target=ctx.identity.rid,
        edge_type=EdgeType.WEBHOOK,
        rid_types=[KoiNetNode, KoiNetEdge]
    )
        
    ctx.kobj_queue.put_kobj(rid=edge_bundle.rid, event_type=EventType.FORGET)
    ctx.kobj_queue.put_kobj(bundle=edge_bundle)

class CoordinatorNodeAssembler(NodeAssembler):
    config = CoordinatorConfig
    knowledge_handlers = [
        *NodeAssembler.knowledge_handlers,
        handshake_handler
    ]


if __name__ == "__main__":
    node = CoordinatorNodeAssembler.create()
    node.server.run()


import structlog
from koi_net.handshaker import Handshaker
from koi_net.protocol.errors import ErrorType
from koi_net.protocol.event import EventType
from rid_lib.types import KoiNetNode
from ..processor.kobj_queue import KobjQueue

log = structlog.stdlib.get_logger()


class ErrorHandler:
    """Handles network errors that may occur during requests."""
    timeout_counter: dict[KoiNetNode, int]
    kobj_queue: KobjQueue
    
    def __init__(
        self, 
        kobj_queue: KobjQueue,
        handshaker: Handshaker
    ):
        self.kobj_queue = kobj_queue
        self.handshaker = handshaker
        self.timeout_counter = {}
        
    def handle_connection_error(self, node: KoiNetNode):
        """Drops nodes after timing out three times."""
        self.timeout_counter.setdefault(node, 0)
        self.timeout_counter[node] += 1
        
        log.debug(f"{node} has timed out {self.timeout_counter[node]} time(s)")
        
        if self.timeout_counter[node] > 3:
            log.debug(f"Exceeded time out limit, forgetting node")
            self.kobj_queue.put_kobj(rid=node, event_type=EventType.FORGET)
            # do something
        
        
    def handle_protocol_error(
        self, 
        error_type: ErrorType, 
        node: KoiNetNode
    ):
        """Attempts handshake when this node is unknown to target."""
        log.info(f"Handling protocol error {error_type} for node {node!r}")
        match error_type:
            case ErrorType.UnknownNode:
                log.info("Peer doesn't know me, attempting handshake...")
                self.handshaker.handshake_with(node)
                
            case ErrorType.InvalidKey: ...
            case ErrorType.InvalidSignature: ...
            case ErrorType.InvalidTarget: ...




import time
import structlog

from koi_net.interfaces.entrypoint import EntryPoint
from .processor.kobj_queue import KobjQueue
from .lifecycle import NodeLifecycle
from .network.resolver import NetworkResolver
from .config import NodeConfig

log = structlog.stdlib.get_logger()


class NodePoller(EntryPoint):
    """Manages polling based event loop for partial nodes."""
    kobj_queue: KobjQueue
    lifecycle: NodeLifecycle
    resolver: NetworkResolver
    config: NodeConfig
    
    def __init__(
        self,
        config: NodeConfig,
        lifecycle: NodeLifecycle,
        kobj_queue: KobjQueue,
        resolver: NetworkResolver,
    ):
        self.kobj_queue = kobj_queue
        self.lifecycle = lifecycle
        self.resolver = resolver
        self.config = config

    def poll(self):
        """Polls neighbors and processes returned events."""
        neighbors = self.resolver.poll_neighbors()
        for node_rid in neighbors:
            for event in neighbors[node_rid]:
                self.kobj_queue.put_kobj(event=event, source=node_rid)

    def run(self):
        """Runs polling event loop."""
        with self.lifecycle.run():
            while True:
                start_time = time.time()
                self.poll()
                elapsed = time.time() - start_time
                sleep_time = self.config.koi_net.polling_interval - elapsed
                if sleep_time > 0:
                    time.sleep(sleep_time)


import queue
import traceback
import time
import structlog

from rid_lib.ext import Cache
from rid_lib.types import KoiNetNode

from koi_net.config import NodeConfig
from koi_net.network.event_queue import EventQueue, QueuedEvent
from koi_net.network.request_handler import RequestHandler
from koi_net.network.poll_event_buffer import PollEventBuffer
from koi_net.protocol.event import Event
from koi_net.protocol.node import NodeProfile, NodeType
from koi_net.interfaces.worker import ThreadWorker, STOP_WORKER

log = structlog.stdlib.get_logger()


class EventProcessingWorker(ThreadWorker):
    event_buffer: dict[KoiNetNode, list[Event]]
    buffer_times: dict[KoiNetNode, float]

    def __init__(
        self,
        event_queue: EventQueue,
        request_handler: RequestHandler,
        config: NodeConfig,
        cache: Cache,
        poll_event_buf: PollEventBuffer
    ):
        self.event_queue = event_queue
        self.request_handler = request_handler
        
        self.config = config
        self.cache = cache
        self.poll_event_buf = poll_event_buf
        
        self.timeout: float = 0.1
        self.max_buf_len: int = 5
        self.max_wait_time: float = 1.0
        
        self.event_buffer = dict()
        self.buffer_times = dict()
        
        super().__init__()
        
    def flush_buffer(self, target: KoiNetNode, buffer: list[Event]):
        try:
            self.request_handler.broadcast_events(target, events=buffer)
        except Exception as e:
            traceback.print_exc()
        
        self.event_buffer[target] = []
        self.buffer_times[target] = None
        
    def decide_event(self, item: QueuedEvent) -> bool:
        node_bundle = self.cache.read(item.target)
        if node_bundle: 
            node_profile = node_bundle.validate_contents(NodeProfile)
            
            if node_profile.node_type == NodeType.FULL:
                return True
        
            elif node_profile.node_type == NodeType.PARTIAL:
                self.poll_event_buf.put(item.target, item.event)
                return False
        
        elif item.target == self.config.koi_net.first_contact.rid:
            return True
        
        else:
            log.warning(f"Couldn't handle event {item.event!r} in queue, node {item.target!r} unknown to me")
            return False
        

    def run(self):
        log.info("Started event worker")
        while True:
            now = time.time()
            try:
                item = self.event_queue.q.get(timeout=self.timeout)
                
                try:
                    if item is STOP_WORKER:
                        log.info(f"Received 'STOP_WORKER' signal, flushing buffer...")
                        for target in self.event_buffer.keys():
                            self.flush_buffer(target, self.event_buffer[target])
                        return
                    
                    log.info(f"Dequeued {item.event!r} -> {item.target!r}")
                    
                    if not self.decide_event(item):
                        continue
                    
                    event_buf = self.event_buffer.setdefault(item.target, [])
                    if not event_buf:
                        self.buffer_times[item.target] = now
                    
                    event_buf.append(item.event)

                    # When new events are dequeued, check buffer for max length
                    if len(event_buf) >= self.max_buf_len:
                        self.flush_buffer(item.target, event_buf)
                finally:
                    self.event_queue.q.task_done()

            except queue.Empty:
                # On timeout, check all buffers for max wait time
                for target, event_buf in self.event_buffer.items():
                    if (len(event_buf) == 0) or (self.buffer_times.get(target) is None):
                        continue
                    if (now - self.buffer_times[target]) >= self.max_wait_time: 
                        self.flush_buffer(target, event_buf)
                    
            except Exception as e:
                traceback.print_exc()


from rid_lib.ext import Cache
from .assembler import NodeAssembler
from .config.base import BaseConfig
from .context import HandlerContext
from .effector import Effector
from .handshaker import Handshaker
from .identity import NodeIdentity
from .processor.kobj_worker import KnowledgeProcessingWorker
from .lifecycle import NodeLifecycle
from .network.error_handler import ErrorHandler
from .network.event_queue import EventQueue
from .network.graph import NetworkGraph
from .network.request_handler import RequestHandler
from .network.resolver import NetworkResolver
from .network.response_handler import ResponseHandler
from .network.poll_event_buffer import PollEventBuffer
from .poller import NodePoller
from .processor.handlers import (
    basic_manifest_handler, 
    basic_network_output_filter, 
    basic_rid_handler, 
    node_contact_handler, 
    edge_negotiation_handler, 
    forget_edge_on_node_deletion, 
    secure_profile_handler
)
from .processor.event_worker import EventProcessingWorker
from .processor.pipeline import KnowledgePipeline
from .processor.kobj_queue import KobjQueue
from .secure import Secure
from .server import NodeServer


# factory functions for components with non standard initializiation

def make_config() -> BaseConfig:
    return BaseConfig.load_from_yaml()

def make_cache(config: BaseConfig) -> Cache:
    return Cache(directory_path=config.koi_net.cache_directory_path)


class BaseNode(NodeAssembler):
    config = lambda: None
    kobj_queue = KobjQueue
    event_queue = EventQueue
    poll_event_buf = PollEventBuffer
    knowledge_handlers = lambda: [
        basic_rid_handler,
        basic_manifest_handler,
        secure_profile_handler,
        edge_negotiation_handler,
        node_contact_handler,
        basic_network_output_filter,
        forget_edge_on_node_deletion
    ]
    cache = make_cache
    identity = NodeIdentity
    graph = NetworkGraph
    secure = Secure
    handshaker = Handshaker
    error_handler = ErrorHandler
    request_handler = RequestHandler
    response_handler = ResponseHandler
    resolver = NetworkResolver
    effector = Effector
    handler_context = HandlerContext
    pipeline = KnowledgePipeline
    kobj_worker = KnowledgeProcessingWorker
    event_worker = EventProcessingWorker
    lifecycle = NodeLifecycle


class FullNode(BaseNode):
    entrypoint = NodeServer

class PartialNode(BaseNode):
    entrypoint = NodePoller


import structlog
from typing import Callable
from enum import StrEnum
from rid_lib.ext import Cache, Bundle
from rid_lib.core import RID, RIDType
from rid_lib.types import KoiNetNode
from .network.resolver import NetworkResolver
from .processor.kobj_queue import KobjQueue
from .identity import NodeIdentity

log = structlog.stdlib.get_logger()


class ActionContext:
    """Provides action handlers access to other subsystems."""
    
    identity: NodeIdentity

    def __init__(
        self,
        identity: NodeIdentity,
    ):
        self.identity = identity
    

class BundleSource(StrEnum):
    CACHE = "CACHE"
    ACTION = "ACTION"

class Effector:
    """Subsystem for dereferencing RIDs."""
    
    cache: Cache
    resolver: NetworkResolver
    kobj_queue: KobjQueue | None
    action_context: ActionContext | None
    _action_table: dict[
        type[RID], 
        Callable[
            [ActionContext, RID], 
            Bundle | None
        ]
    ] = dict()
    
    def __init__(
        self, 
        cache: Cache,
        resolver: NetworkResolver,
        kobj_queue: KobjQueue,
        identity: NodeIdentity
    ):
        self.cache = cache
        self.resolver = resolver
        self.kobj_queue = kobj_queue
        self.action_context = ActionContext(identity)
        self._action_table = self.__class__._action_table.copy()
    
    @classmethod
    def register_default_action(cls, rid_type: RIDType):
        def decorator(func: Callable) -> Callable:
            cls._action_table[rid_type] = func
            return func
        return decorator
        
    def register_action(self, rid_type: RIDType):
        """Registers a new dereference action for an RID type.
        
        Example:
            This function should be used as a decorator on an action function::
            
                @node.register_action(KoiNetNode)
                def deref_koi_net_node(ctx: ActionContext, rid: KoiNetNode):
                    # return a Bundle or None
                    return
        """
        def decorator(func: Callable) -> Callable:
            self._action_table[rid_type] = func
            return func
        return decorator
    
    def _try_cache(self, rid: RID) -> tuple[Bundle, BundleSource] | None:
        bundle = self.cache.read(rid)
        
        if bundle:
            log.debug("Cache hit")
            return bundle, BundleSource.CACHE
        else:
            log.debug("Cache miss")
            return None
                    
    def _try_action(self, rid: RID) -> tuple[Bundle, BundleSource] | None:
        if type(rid) not in self._action_table:
            log.debug("No action available")
            return None
        
        log.debug("Action available")
        func = self._action_table[type(rid)]
        bundle = func(
            ctx=self.action_context, 
            rid=rid
        )
        
        if bundle:
            log.debug("Action hit")
            return bundle, BundleSource.ACTION
        else:
            log.debug("Action miss")
            return None

        
    def _try_network(self, rid: RID) -> tuple[Bundle, KoiNetNode] | None:
        bundle, source = self.resolver.fetch_remote_bundle(rid)
        
        if bundle:
            log.debug("Network hit")
            return bundle, source
        else:
            log.debug("Network miss")
            return None
        
    
    def deref(
        self, 
        rid: RID,
        refresh_cache: bool = False,
        use_network: bool = False,
        handle_result: bool = True
    ) -> Bundle | None:
        """Dereferences an RID.
        
        Attempts to dereference an RID by (in order) reading the cache, 
        calling a bound action, or fetching from other nodes in the 
        newtork.
        
        Args:
            rid: RID to dereference
            refresh_cache: skips cache read when `True` 
            use_network: enables fetching from other nodes when `True`
            handle_result: handles resulting bundle with knowledge pipeline when `True`
        """
        
        log.debug(f"Dereferencing {rid!r}")
        
        bundle, source = (
            # if `refresh_cache`, skip try cache
            not refresh_cache and self._try_cache(rid) or 
            self._try_action(rid) or
            use_network and self._try_network(rid) or
            # if not found, bundle and source set to None
            (None, None) 
        )
        
        if (
            handle_result 
            and bundle is not None 
            and source != BundleSource.CACHE
        ):            
            self.kobj_queue.put_kobj(
                bundle=bundle, 
                source=source if type(source) is KoiNetNode else None
            )

            # TODO: refactor for general solution, param to write through to cache before continuing
            # like `self.processor.kobj_queue.join()``

        return bundle


# import sentry_sdk
# from sentry_sdk.integrations.logging import LoggingIntegration

# sentry_sdk.init(
#     dsn="https://7bbafef3c7dbd652506db3cb2aca9f98@o4510149352357888.ingest.us.sentry.io/4510149355765760",
#     # Add data like request headers and IP for users,
#     # see https://docs.sentry.io/platforms/python/data-management/data-collected/ for more info
#     send_default_pii=True,
#     enable_logs=True,
#     integrations=[
#         LoggingIntegration(sentry_logs_level=None)
#     ]
# )


import os
import typer
from typing import Callable
from rich.console import Console
from rich.table import Table

from importlib.metadata import entry_points

from koi_net.cli.models import KoiNetworkConfig
from koi_net.core import NodeInterface
import shutil

app = typer.Typer()
console = Console()

installed_nodes = entry_points(group='koi_net.node')

net_config = KoiNetworkConfig.load_from_yaml()

@app.command()
def list_node_types():
    table = Table(title="installed node types")
    table.add_column("name", style="cyan")
    table.add_column("module", style="magenta")

    for node in installed_nodes:
        table.add_row(node.name, node.module)
    console.print(table)
    
@app.command()
def list_nodes():
    table = Table(title="created nodes")
    table.add_column("name", style="cyan")
    table.add_column("rid", style="magenta")

    for dir in os.listdir('.'):
        if not os.path.isdir(dir):
            continue
        for file in os.listdir(dir):
            file_path = os.path.join(dir, file)
            if not (os.path.isfile(file_path) and file == "config.yaml"):
                continue
            
            print(os.getcwd())
            os.chdir(dir)
            print(os.getcwd())
                        
            node_type = net_config.nodes.get(dir)
            
            ep = list(installed_nodes.select(name=node_type))[0]
            create_node: Callable[[], NodeInterface] = ep.load()
            
            node = create_node()
            
            print(ep)
            print(dir)
            print(node.identity.rid)
            
            table.add_row(dir, str(node.identity.rid))
            
            os.chdir('..')
            print(os.getcwd())
    
    console.print(table)

@app.command()
def create(type: str, name: str):
    # if name not in installed_nodes:
    #     console.print(f"[bold red]Error:[/bold red] node type '{name}' doesn't exist")
    #     raise typer.Exit(code=1)

    eps = installed_nodes.select(name=type)
    if eps:
        ep = list(eps)[0]
    
    os.mkdir(name)
    os.chdir(name)
    
    ep.load()
    
    os.chdir('..')
    
    net_config.nodes[name] = type
    net_config.save_to_yaml()
    
@app.command()
def remove(name: str):
    shutil.rmtree(name)
    net_config.nodes.pop(name, None)
    net_config.save_to_yaml()
    
@app.command()
def start(name: str):
    os.chdir(name)
    node_type = net_config.nodes.get(name)
    ep = list(installed_nodes.select(name=node_type))[0]
    create_node: Callable[[], NodeInterface] = ep.load()
    
    create_node().server.run()


import structlog
from contextlib import contextmanager, asynccontextmanager

from rid_lib.ext import Bundle, Cache
from rid_lib.types import KoiNetNode

from .handshaker import Handshaker
from .network.request_handler import RequestHandler
from .processor.kobj_worker import KnowledgeProcessingWorker
from .network.event_queue import EventQueue
from .processor.event_worker import EventProcessingWorker
from .protocol.api_models import ErrorResponse
from .interfaces.worker import STOP_WORKER
from .config import NodeConfig
from .processor.kobj_queue import KobjQueue
from .network.graph import NetworkGraph
from .identity import NodeIdentity

log = structlog.stdlib.get_logger()


class NodeLifecycle:
    """Manages node startup and shutdown processes."""
    
    config: NodeConfig
    identity: NodeIdentity
    graph: NetworkGraph
    kobj_queue: KobjQueue
    kobj_worker: KnowledgeProcessingWorker
    event_queue: EventQueue
    event_worker: EventProcessingWorker
    cache: Cache
    handshaker: Handshaker
    request_handler: RequestHandler
    
    def __init__(
        self,
        config: NodeConfig,
        identity: NodeIdentity,
        graph: NetworkGraph,
        kobj_queue: KobjQueue,
        kobj_worker: KnowledgeProcessingWorker,
        event_queue: EventQueue,
        event_worker: EventProcessingWorker,
        cache: Cache,
        handshaker: Handshaker,
        request_handler: RequestHandler
    ):
        self.config = config
        self.identity = identity
        self.graph = graph
        self.kobj_queue = kobj_queue
        self.kobj_worker = kobj_worker
        self.event_queue = event_queue
        self.event_worker = event_worker
        self.cache = cache
        self.handshaker = handshaker
        self.request_handler = request_handler
        
    @contextmanager
    def run(self):
        """Synchronous context manager for node startup and shutdown."""
        try:
            log.info("Starting node lifecycle...")
            self.start()
            yield
        except KeyboardInterrupt:
            log.info("Keyboard interrupt!")
        finally:
            log.info("Stopping node lifecycle...")
            self.stop()

    @asynccontextmanager
    async def async_run(self):
        """Asynchronous context manager for node startup and shutdown."""
        try:
            log.info("Starting async node lifecycle...")
            self.start()
            yield
        except KeyboardInterrupt:
            log.info("Keyboard interrupt!")
        finally:
            log.info("Stopping async node lifecycle...")
            self.stop()
    
    def start(self):
        """Starts a node.
        
        Starts the processor thread (if enabled). Generates network 
        graph from nodes and edges in cache. Processes any state changes 
        of node bundle. Initiates handshake with first contact if node 
        doesn't have any neighbors. Catches up with coordinator state.
        """
        log.info("Starting processor worker thread")
        
        self.kobj_worker.thread.start()
        self.event_worker.thread.start()
        self.graph.generate()
        
        # refresh to reflect changes (if any) in config.yaml
        
        self.kobj_queue.put_kobj(bundle=Bundle.generate(
            rid=self.identity.rid,
            contents=self.identity.profile.model_dump()
        ))
        
        log.debug("Waiting for kobj queue to empty")
        self.kobj_queue.q.join()
        
        coordinators = self.graph.get_neighbors(direction="in", allowed_type=KoiNetNode)
        
        if len(coordinators) > 0:
            for coordinator in coordinators:
                payload = self.request_handler.fetch_manifests(
                    node=coordinator,
                    rid_types=[KoiNetNode]
                )
                if type(payload) is ErrorResponse:
                    continue
                
                for manifest in payload.manifests:
                    self.kobj_queue.put_kobj(
                        manifest=manifest,
                        source=coordinator
                    )
                    
        elif self.config.koi_net.first_contact.rid:
            log.debug(f"I don't have any edges with coordinators, reaching out to first contact {self.config.koi_net.first_contact.rid!r}")
            
            self.handshaker.handshake_with(self.config.koi_net.first_contact.rid)
        

    def stop(self):
        """Stops a node.
        
        Finishes processing knowledge object queue.
        """        
        log.info(f"Waiting for kobj queue to empty ({self.kobj_queue.q.unfinished_tasks} tasks remaining)")
        
        self.kobj_queue.q.put(STOP_WORKER)
        self.event_queue.q.put(STOP_WORKER)


import inspect
from typing import Protocol
from dataclasses import make_dataclass

import structlog

from .interfaces.entrypoint import EntryPoint

log = structlog.stdlib.get_logger()


class BuildOrderer(type):
    def __new__(cls, name: str, bases: tuple, dct: dict[str]):
        """Sets `cls._build_order` from component order in class definition."""
        cls = super().__new__(cls, name, bases, dct)
        
        if "_build_order" not in dct:
            components = {}
            # adds components from base classes (including cls)
            for base in reversed(inspect.getmro(cls)[:-1]):
                for k, v in vars(base).items():
                    # excludes built in and private attributes
                    if not k.startswith("_"):
                        components[k] = v
                        
            # recipe list constructed from names of non-None components
            cls._build_order = [
                name for name, _type in components.items()
                if _type is not None
            ]
            
        return cls

class NodeContainer(Protocol):
    entrypoint = EntryPoint

class NodeAssembler(metaclass=BuildOrderer):    
    def __new__(self) -> NodeContainer:
        return self._build()
    
    @classmethod
    def _build(cls) -> NodeContainer:
        components = {}
        for comp_name in cls._build_order:            
            comp_factory = getattr(cls, comp_name, None)
            
            if comp_factory is None:
                raise Exception(f"Couldn't find factory for component '{comp_name}'")
            
            sig = inspect.signature(comp_factory)
            
            required_comps = []
            for name, param in sig.parameters.items():
                required_comps.append((name, param.annotation))
            
            if len(required_comps) == 0:
                s = comp_name
            else:
                s = f"{comp_name} -> {', '.join([name for name, _type in required_comps])}"
            
            print(s.replace("graph", "_graph"), end=";\n")
            
            dependencies = {}
            for req_comp_name, req_comp_type in required_comps:
                if req_comp_name not in components:
                    raise Exception(f"Couldn't find required component '{req_comp_name}'")
                    
                dependencies[req_comp_name] = components[req_comp_name]
                
            components[comp_name] = comp_factory(**dependencies)
        
        NodeContainer = make_dataclass(
            cls_name="NodeContainer",
            fields=[
                (name, type(component)) 
                for name, component
                in components.items()
            ],
            frozen=True
        )
        
        return NodeContainer(**components)



from rid_lib.ext import Cache

from koi_net.effector import Effector
from koi_net.network.resolver import NetworkResolver
from .config import NodeConfig
from .network.graph import NetworkGraph
from .network.event_queue import EventQueue
from .network.request_handler import RequestHandler
from .identity import NodeIdentity
from .processor.kobj_queue import KobjQueue


class HandlerContext:
    """Provides knowledge handlers access to other subsystems."""
    
    identity: NodeIdentity
    config: NodeConfig
    cache: Cache
    event_queue: EventQueue
    kobj_queue: KobjQueue
    graph: NetworkGraph
    request_handler: RequestHandler
    resolver: NetworkResolver
    effector: Effector
    
    def __init__(
        self,
        identity: NodeIdentity,
        config: NodeConfig,
        cache: Cache,
        event_queue: EventQueue,
        kobj_queue: KobjQueue,
        graph: NetworkGraph,
        request_handler: RequestHandler,
        resolver: NetworkResolver,
        effector: Effector
    ):
        self.identity = identity
        self.config = config
        self.cache = cache
        self.event_queue = event_queue
        self.kobj_queue = kobj_queue
        self.graph = graph
        self.request_handler = request_handler
        self.resolver = resolver
        self.effector = effector


import queue
import traceback
import structlog

from .pipeline import KnowledgePipeline
from .kobj_queue import KobjQueue
from koi_net.interfaces.worker import ThreadWorker, STOP_WORKER

log = structlog.stdlib.get_logger()


class KnowledgeProcessingWorker(ThreadWorker):
    def __init__(
        self,
        kobj_queue: KobjQueue,
        pipeline: KnowledgePipeline
    ):
        self.kobj_queue = kobj_queue
        self.pipeline = pipeline
        self.timeout: float = 0.1

        super().__init__()
        
    def run(self):
        log.info("Started kobj worker")
        while True:
            try:
                item = self.kobj_queue.q.get(timeout=self.timeout)
                try:
                    if item is STOP_WORKER:
                        log.info("Received 'STOP_WORKER' signal, shutting down...")
                        return
                    
                    log.info(f"Dequeued {item!r}")
                    
                    self.pipeline.process(item)
                finally:
                    self.kobj_queue.q.task_done()
                    
            except queue.Empty:
                pass
            
            except Exception as e:
                traceback.print_exc()


from datetime import datetime
import logging
from logging.handlers import RotatingFileHandler
import colorama
import structlog
import sys
# from sentry_sdk import logger as sentry_logger


def my_processor(_, __, event: dict):
    # print(_, __, event)
    event["path"] = event["module"] + "." + event["func_name"]
    return event

# def sentry_processor(_, method, event: dict):
#     print(event)
#     if method == "critical":
#         sentry_logger.fatal(
#             event["event"],
#             attributes=event
#         )
#     elif method == "info":
#         sentry_logger.info(
#             event["event"],
#             attributes=event
#         )
#     elif method == "debug":
#         sentry_logger.debug(
#             event["event"],
#             attributes=event
#         )
#     return event

console_renderer = structlog.dev.ConsoleRenderer(
    columns=[
        # Render the timestamp without the key name in yellow.
        structlog.dev.Column(
            "timestamp",
            structlog.dev.KeyValueColumnFormatter(
                key_style=None,
                value_style=colorama.Style.DIM,
                reset_style=colorama.Style.RESET_ALL,
                value_repr=lambda t: datetime.fromisoformat(t).strftime("%Y-%m-%d %H:%M:%S"),
            ),
        ),
        structlog.dev.Column(
            "level",
            structlog.dev.LogLevelColumnFormatter(
                level_styles={
                    level: colorama.Style.BRIGHT + color
                    for level, color in {
                        "critical": colorama.Fore.RED,
                        "exception": colorama.Fore.RED,
                        "error": colorama.Fore.RED,
                        "warn": colorama.Fore.YELLOW,
                        "warning": colorama.Fore.YELLOW,
                        "info": colorama.Fore.GREEN,
                        "debug": colorama.Fore.GREEN,
                        "notset": colorama.Back.RED,
                    }.items()
                },
                reset_style=colorama.Style.RESET_ALL,
                width=9
            )
        ),
        # Render the event without the key name in bright magenta.
        
        # Default formatter for all keys not explicitly mentioned. The key is
        # cyan, the value is green.
        structlog.dev.Column(
            "path",
            structlog.dev.KeyValueColumnFormatter(
                key_style=None,
                value_style=colorama.Fore.MAGENTA,
                reset_style=colorama.Style.RESET_ALL,
                value_repr=str,
                width=30
            ),
        ),
        # structlog.dev.Column(
        #     "func_name",
        #     structlog.dev.KeyValueColumnFormatter(
        #         key_style=None,
        #         value_style=colorama.Fore.MAGENTA,
        #         reset_style=colorama.Style.RESET_ALL,
        #         value_repr=str,
        #         prefix="(",
        #         postfix=")",
        #         width=15
        #     ),
        # ),
        structlog.dev.Column(
            "event",
            structlog.dev.KeyValueColumnFormatter(
                key_style=None,
                value_style=colorama.Fore.WHITE,
                reset_style=colorama.Style.RESET_ALL,
                value_repr=str,
                width=30
            ),
        ),
        structlog.dev.Column(
            "",
            structlog.dev.KeyValueColumnFormatter(
                key_style=colorama.Fore.BLUE,
                value_style=colorama.Fore.GREEN,
                reset_style=colorama.Style.RESET_ALL,
                value_repr=str,
            ),
        )
    ]
)

structlog.configure(
    processors=[
        # If log level is too low, abort pipeline and throw away log entry.
        structlog.stdlib.filter_by_level,
        # Add the name of the logger to event dict.
        structlog.stdlib.add_logger_name,
        # Add log level to event dict.
        structlog.stdlib.add_log_level,
        # Perform %-style formatting.
        structlog.stdlib.PositionalArgumentsFormatter(),
        # Add a timestamp in ISO 8601 format.
        structlog.processors.TimeStamper(fmt="iso"),
        # If the "stack_info" key in the event dict is true, remove it and
        # render the current stack trace in the "stack" key.
        structlog.processors.StackInfoRenderer(),
        # If the "exc_info" key in the event dict is either true or a
        # sys.exc_info() tuple, remove "exc_info" and render the exception
        # with traceback into the "exception" key.
        # structlog.processors.format_exc_info,
        # If some value is in bytes, decode it to a Unicode str.
        structlog.processors.UnicodeDecoder(),
        # Add callsite parameters.
        structlog.processors.CallsiteParameterAdder(
            {
                structlog.processors.CallsiteParameter.MODULE,
                structlog.processors.CallsiteParameter.FUNC_NAME,
                # structlog.processors.CallsiteParameter.LINENO,
            }
        ),
        my_processor,
        # Render the final event dict as JSON.
        # sentry_processor,
        console_renderer
        # structlog.processors.JSONRenderer()
        
    ],
    # `wrapper_class` is the bound logger that you get back from
    # get_logger(). This one imitates the API of `logging.Logger`.
    wrapper_class=structlog.stdlib.BoundLogger,
    # `logger_factory` is used to create wrapped loggers that are used for
    # OUTPUT. This one returns a `logging.Logger`. The final value (a JSON
    # string) from the final processor (`JSONRenderer`) will be passed to
    # the method of the same name as that you've called on the bound logger.
    logger_factory=structlog.stdlib.LoggerFactory(),
    # Effectively freeze configuration after creating the first bound
    # logger.
    cache_logger_on_first_use=True,
)

file_handler = RotatingFileHandler(
    filename="app.log",
    maxBytes=10 * 1024 * 1024,
    backupCount=5,
    encoding="utf-8"
)

logging.basicConfig(
    format="%(message)s",
    stream=sys.stdout,
    level=logging.INFO,
)

# log = structlog.stdlib.get_logger()


import structlog
import httpx
from rid_lib import RID
from rid_lib.ext import Cache
from rid_lib.types.koi_net_node import KoiNetNode

from koi_net.protocol.model_map import API_MODEL_MAP

from ..identity import NodeIdentity
from ..protocol.api_models import (
    RidsPayload,
    ManifestsPayload,
    BundlesPayload,
    EventsPayload,
    FetchRids,
    FetchManifests,
    FetchBundles,
    PollEvents,
    RequestModels,
    ResponseModels,
    ErrorResponse
)
from ..protocol.envelope import SignedEnvelope
from ..protocol.consts import (
    BROADCAST_EVENTS_PATH,
    POLL_EVENTS_PATH,
    FETCH_RIDS_PATH,
    FETCH_MANIFESTS_PATH,
    FETCH_BUNDLES_PATH
)
from ..protocol.node import NodeProfile, NodeType
from ..secure import Secure
from .error_handler import ErrorHandler

log = structlog.stdlib.get_logger()


# Custom error types for request handling
class SelfRequestError(Exception):
    """Raised when a node tries to request itself."""
    pass

class PartialNodeQueryError(Exception):
    """Raised when attempting to query a partial node."""
    pass

class NodeNotFoundError(Exception):
    """Raised when a node URL cannot be found."""
    pass

class UnknownPathError(Exception):
    """Raised when an unknown path is requested."""
    pass

class RequestHandler:
    """Handles making requests to other KOI nodes."""
    
    cache: Cache
    identity: NodeIdentity
    secure: Secure
    error_handler: ErrorHandler
    
    def __init__(
        self, 
        cache: Cache,
        identity: NodeIdentity,
        secure: Secure,
        error_handler: ErrorHandler
    ):
        self.cache = cache
        self.identity = identity
        self.secure = secure
        self.error_handler = error_handler
    
    def get_url(self, node_rid: KoiNetNode) -> str:
        """Retrieves URL of a node from its RID."""
        
        log.debug(f"Getting URL for {node_rid!r}")
        node_url = None
        
        if node_rid == self.identity.rid:
            raise SelfRequestError("Don't talk to yourself")
        
        node_bundle = self.cache.read(node_rid)
                
        if node_bundle:
            node_profile = node_bundle.validate_contents(NodeProfile)
            log.debug(f"Found node profile: {node_profile}")
            if node_profile.node_type != NodeType.FULL:
                raise PartialNodeQueryError("Can't query partial node")
            node_url = node_profile.base_url
        
        else:
            if node_rid == self.identity.config.koi_net.first_contact.rid:
                log.debug("Found URL of first contact")
                node_url = self.identity.config.koi_net.first_contact.url
        
        if not node_url:
            raise NodeNotFoundError("Node not found")
        
        log.debug(f"Resolved {node_rid!r} to {node_url}")
        return node_url
    
    def make_request(
        self,
        node: KoiNetNode,
        path: str, 
        request: RequestModels,
    ) -> ResponseModels | None:
        """Makes a request to a node."""
        url = self.get_url(node) + path
        log.info(f"Making request to {url}")
    
        signed_envelope = self.secure.create_envelope(
            payload=request,
            target=node
        )
        
        try:
            result = httpx.post(
                url, 
                data=signed_envelope.model_dump_json(exclude_none=True)
            )
        except httpx.ConnectError as err:
            log.debug("Failed to connect")
            self.error_handler.handle_connection_error(node)
            raise err
        
        if result.status_code != 200:
            resp = ErrorResponse.model_validate_json(result.text)
            self.error_handler.handle_protocol_error(resp.error, node)
            return resp
        
        resp_env_model = API_MODEL_MAP[path].response_envelope
        if resp_env_model is None:
            return
        
        resp_envelope = resp_env_model.model_validate_json(result.text)
        self.secure.validate_envelope(resp_envelope)
        
        return resp_envelope.payload
    
    def broadcast_events(
        self, 
        node: RID, 
        req: EventsPayload | None = None,
        **kwargs
    ) -> None:
        """Broadcasts events to a node.
        
        Pass `EventsPayload` object, or see `protocol.api_models.EventsPayload` for available kwargs.
        """
        request = req or EventsPayload.model_validate(kwargs)
        self.make_request(node, BROADCAST_EVENTS_PATH, request)
        log.info(f"Broadcasted {len(request.events)} event(s) to {node!r}")
        
    def poll_events(
        self, 
        node: RID, 
        req: PollEvents | None = None,
        **kwargs
    ) -> EventsPayload:
        """Polls events from a node.
        
        Pass `PollEvents` object as `req` or fields as kwargs.
        """
        request = req or PollEvents.model_validate(kwargs)
        resp = self.make_request(node, POLL_EVENTS_PATH, request)
        if type(resp) != ErrorResponse:
            log.info(f"Polled {len(resp.events)} events from {node!r}")
        return resp
        
    def fetch_rids(
        self, 
        node: RID, 
        req: FetchRids | None = None,
        **kwargs
    ) -> RidsPayload:
        """Fetches RIDs from a node.
        
        Pass `FetchRids` object as `req` or fields as kwargs.
        """
        request = req or FetchRids.model_validate(kwargs)
        resp = self.make_request(node, FETCH_RIDS_PATH, request)
        if type(resp) != ErrorResponse:
            log.info(f"Fetched {len(resp.rids)} RID(s) from {node!r}")
        return resp
                
    def fetch_manifests(
        self, 
        node: RID, 
        req: FetchManifests | None = None,
        **kwargs
    ) -> ManifestsPayload:
        """Fetches manifests from a node.
        
        Pass `FetchManifests` object as `req` or fields as kwargs.
        """
        request = req or FetchManifests.model_validate(kwargs)
        resp = self.make_request(node, FETCH_MANIFESTS_PATH, request)
        if type(resp) != ErrorResponse:
            log.info(f"Fetched {len(resp.manifests)} manifest(s) from {node!r}")
        return resp
                
    def fetch_bundles(
        self, 
        node: RID, 
        req: FetchBundles | None = None,
        **kwargs
    ) -> BundlesPayload:
        """Fetches bundles from a node.
        
        Pass `FetchBundles` object as `req` or fields as kwargs.
        """
        request = req or FetchBundles.model_validate(kwargs)
        resp = self.make_request(node, FETCH_BUNDLES_PATH, request)
        if type(resp) != ErrorResponse:
            log.info(f"Fetched {len(resp.bundles)} bundle(s) from {node!r}")
        return resp


from pydantic import BaseModel, Field, PrivateAttr
from ruamel.yaml import YAML


class KoiNetworkConfig(BaseModel):
    nodes: dict[str, str] = Field(default_factory=dict)
    _file_path: str = PrivateAttr(default="koi-net-config.yaml")

    @classmethod
    def load_from_yaml(
        cls, 
        file_path: str = "koi-net-config.yaml", 
    ):
        yaml = YAML()
        
        try:
            with open(file_path, "r") as f:
                file_content = f.read()
            config_data = yaml.load(file_content)
            config = cls.model_validate(config_data)
            
        except FileNotFoundError:
            config = cls()
        
        config._file_path = file_path
        config.save_to_yaml()
        return config
    
    def save_to_yaml(self):
        yaml = YAML()
        
        with open(self._file_path, "w") as f:
            try:
                config_data = self.model_dump(mode="json")
                yaml.dump(config_data, f)
            except Exception as e:
                if self._file_content:
                    f.seek(0)
                    f.truncate()
                    f.write(self._file_content)
                raise e


import os
from rid_lib import RIDType
from ruamel.yaml import YAML
from pydantic import BaseModel, Field, PrivateAttr
from dotenv import load_dotenv
from rid_lib.ext.utils import sha256_hash
from rid_lib.types import KoiNetNode
from .protocol.secure import PrivateKey
from .protocol.node import NodeProfile, NodeType


class ServerConfig(BaseModel):
    """Config for the node server (full node only)."""
    
    host: str = "127.0.0.1"
    port: int = 8000
    path: str | None = "/koi-net"
    
    @property
    def url(self) -> str:
        return f"http://{self.host}:{self.port}{self.path or ''}"

class NodeContact(BaseModel):
    rid: KoiNetNode | None = None
    url: str | None = None

class KoiNetConfig(BaseModel):
    """Config for KOI-net."""
    
    node_name: str
    node_rid: KoiNetNode | None = None
    node_profile: NodeProfile
    
    rid_types_of_interest: list[RIDType] = Field(
        default_factory=lambda: [KoiNetNode])
    
    cache_directory_path: str = ".rid_cache"
    event_queues_path: str = "event_queues.json"
    private_key_pem_path: str = "priv_key.pem"
    polling_interval: int = 5
    
    first_contact: NodeContact = Field(default_factory=NodeContact)
    
    _priv_key: PrivateKey | None = PrivateAttr(default=None)

class EnvConfig(BaseModel):
    """Config for environment variables.
    
    Values set in the config are the variables names, and are loaded
    from the environment at runtime. For example, if the config YAML
    sets `priv_key_password: PRIV_KEY_PASSWORD` accessing 
    `priv_key_password` would retrieve the value of `PRIV_KEY_PASSWORD`
    from the environment.
    """
    
    priv_key_password: str | None = "PRIV_KEY_PASSWORD"
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        load_dotenv()
    
    def __getattribute__(self, name):
        value = super().__getattribute__(name)
        if name in type(self).model_fields:
            env_val = os.getenv(value)
            if env_val is None:
                raise ValueError(f"Required environment variable {value} not set")
            return env_val
        return value

class NodeConfig(BaseModel):
    """Base configuration class for all nodes.
    
    Designed to be extensible for custom node implementations. Classes
    inheriting from `NodeConfig` may add additional config groups.
    """
    
    server: ServerConfig = Field(default_factory=ServerConfig)
    koi_net: KoiNetConfig
    env: EnvConfig = Field(default_factory=EnvConfig)
    
    _file_path: str = PrivateAttr(default="config.yaml")
    _file_content: str | None = PrivateAttr(default=None)
    
    @classmethod
    def load_from_yaml(
        cls, 
        file_path: str = "config.yaml", 
        generate_missing: bool = True
    ):
        """Loads config state from YAML file.
        
        Defaults to `config.yaml`. If `generate_missing` is set to 
        `True`, a private key and RID will be generated if not already
        present in the config.
        """
        yaml = YAML()
        
        try:
            with open(file_path, "r") as f:
                file_content = f.read()
            config_data = yaml.load(file_content)
            config = cls.model_validate(config_data)
            config._file_content = file_content
            
        except FileNotFoundError:
            # empty_fields = {}
            # for name, field in cls.model_fields.items():
                
            #     if field.default is None or field.default_factory is None:
            # print(empty_fields)
            config = cls()
            
            
        config._file_path = file_path
        
        if generate_missing:
            if not config.koi_net.node_rid:
                priv_key = PrivateKey.generate()
                pub_key = priv_key.public_key()
                
                config.koi_net.node_rid = KoiNetNode(
                    config.koi_net.node_name,
                    sha256_hash(pub_key.to_der())
                )
                
                with open(config.koi_net.private_key_pem_path, "w") as f:
                    f.write(
                        priv_key.to_pem(config.env.priv_key_password)
                    )
                
                config.koi_net.node_profile.public_key = pub_key.to_der()
            
            if config.koi_net.node_profile.node_type == NodeType.FULL:
                config.koi_net.node_profile.base_url = (
                    config.koi_net.node_profile.base_url or config.server.url
                )
                
            config.save_to_yaml()
                    
        return config
    
    def save_to_yaml(self):
        """Saves config state to YAML file.
        
        File path is set by `load_from_yaml` class method.
        """
        
        yaml = YAML()
        
        with open(self._file_path, "w") as f:
            try:
                config_data = self.model_dump(mode="json")
                yaml.dump(config_data, f)
            except Exception as e:
                if self._file_content:
                    f.seek(0)
                    f.truncate()
                    f.write(self._file_content)
                raise e
                



import structlog
from queue import Queue

from rid_lib.types import KoiNetNode
from pydantic import BaseModel

from ..protocol.event import Event

log = structlog.stdlib.get_logger()


class QueuedEvent(BaseModel):
    event: Event
    target: KoiNetNode

class EventQueue:
    """Handles out going network event queues."""
    q: Queue[QueuedEvent]
    
    def __init__(self):
        self.q = Queue()
    
    def push_event_to(self, event: Event, target: KoiNetNode):
        """Pushes event to queue of specified node.
        
        Event will be sent to webhook or poll queue depending on the 
        node type and edge type of the specified node.
        """
        
        self.q.put(QueuedEvent(target=target, event=event))
    


from rid_lib.types import KoiNetNode

from koi_net.protocol.event import Event


class PollEventBuffer:
    buffers: dict[KoiNetNode, list[Event]]
    
    def __init__(self):
        self.buffers = dict()
        
    def put(self, node: KoiNetNode, event: Event):
        event_buf = self.buffers.setdefault(node, [])
        event_buf.append(event)
        
    def flush(self, node: KoiNetNode, limit: int = 0):
        event_buf = self.buffers.get(node, [])
        
        if limit and len(event_buf) > limit:
            to_return = event_buf[:limit]
            self.buffers[node] = event_buf[limit:]
        else:
            to_return = event_buf.copy()
            self.buffers[node] = []
        
        return to_return


import logging
from pydantic import Field
from rich.logging import RichHandler
from koi_net.core import NodeAssembler
from koi_net.protocol.node import NodeProfile, NodeType
from koi_net.config import NodeConfig, KoiNetConfig

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    handlers=[RichHandler()]
)

logging.getLogger("koi_net").setLevel(logging.DEBUG)
logger = logging.getLogger(__name__)


class PartialNodeConfig(NodeConfig):
    koi_net: KoiNetConfig = Field(default_factory = lambda:
        KoiNetConfig(
            node_name="partial",
            node_profile=NodeProfile(
                node_type=NodeType.PARTIAL
            )
        )
    )

class PartialNodeAssembler(NodeAssembler):
    config = PartialNodeConfig


if __name__ == "__main__":
    node = PartialNodeAssembler.create()
    node.poller.run()